from pathlib import Path

from safeds.data.tabular.containers import Table

from safeds_datasets.tabular.containers import TableWithDescriptions

_path = Path(__file__).parent / "data" / "house_sales.csv"


def load_house_sales() -> TableWithDescriptions:
    """
    Load the "House Sales" dataset.

    Returns
    -------
    house_sales:
        The "House Sales" dataset.
    """
    return TableWithDescriptions(
        Table.from_csv_file(_path),
        column_descriptions={
            "id": "A unique identifier",
            "year": "Year of sale",
            "month": "Month of sale",
            "day": "Day of sale",
            "zipcode": "Zipcode",
            "latitude": "Latitude",
            "longitude": "Longitude",
            "sqft_lot": "Lot area in square feet",
            "sqft_living": "Interior living space in square feet",
            "sqft_above": "Interior living space above ground in square feet",
            "sqft_basement": "Interior living space below ground in square feet",
            "floors": "Number of floors",
            "bedrooms": "Number of bedrooms",
            "bathrooms": (
                "Number of bathrooms. "
                "Fractional values indicate that components (toilet/sink/shower/bathtub) are missing."
            ),
            "waterfront": "Whether the building overlooks a waterfront (0 = no, 1 = yes)",
            "view": "Rating of the view (1 to 5, higher is better)",
            "condition": "Rating of the condition of the house (1 to 5, higher is better)",
            "grade": "Rating of building construction and design (1 to 13, higher is better)",
            "year_built": "Year the house was built",
            "year_renovated": (
                "Year the house was last renovated. A value of 0 indicates that it was never renovated."
            ),
            "sqft_lot_15nn": "Lot area of the 15 nearest neighbors in square feet",
            "sqft_living_15nn": "Interior living space of the 15 nearest neighbors in square feet",
            "price": "Price the house sold for in USD",
        },
    )



from safeds._validation import _check_columns_exist
from safeds.data.tabular.containers import Table


class TableWithDescriptions:
    """
    A `Table` with descriptions for its columns.

    Parameters
    ----------
    table:
        The table.
    column_descriptions:
        A dictionary mapping column names to their descriptions.

    Raises
    ------
    ColumnNotFoundError
        If a column name in `descriptions` does not exist in `table`.
    """

    def __init__(self, table: Table, column_descriptions: dict[str, str]) -> None:
        # Check that all column names in `descriptions` exist in `table`
        _check_columns_exist(table, list(column_descriptions.keys()))

        self._data = table
        self._descriptions = column_descriptions

    @property
    def data(self) -> Table:
        """The data."""
        return self._data

    @property
    def column_descriptions(self) -> Table:
        """
        Return a `Table` contain the name of a column and its description.

        The name is stored in a column called `"Name"` and the description in a column called `"Description"`.
        """
        return Table(
            {
                "Name": self._data.column_names,
                "Description": [self.get_column_description(column_name) for column_name in self._data.column_names],
            },
        )

    def get_column_description(self, column_name: str) -> str:
        """
        Get the description of a column. If no description exists, an empty string is returned.

        Parameters
        ----------
        column_name:
            The name of the column.

        Returns
        -------
        column_description:
            The description of the column.

        Raises
        ------
        ColumnNotFoundError
            If no column with the given name exists.
        """
        _check_columns_exist(self._data, [column_name])

        return self._descriptions.get(column_name, "")



from pathlib import Path

from safeds.data.tabular.containers import Table

from safeds_datasets.tabular.containers import TableWithDescriptions

_path = Path(__file__).parent / "data" / "titanic.csv"


def load_titanic() -> TableWithDescriptions:
    """
    Load the "Titanic" dataset.

    Returns
    -------
    titanic:
        The "Titanic" dataset.
    """
    return TableWithDescriptions(
        Table.from_csv_file(_path),
        column_descriptions={
            "id": "A unique identifier",
            "name": "Name of the passenger",
            "sex": "Sex of the passenger",
            "age": "Age of the passenger at the time of the accident",
            "siblings_spouses": "Number of siblings or spouses aboard",
            "parents_children": "Number of parents or children aboard",
            "ticket": "Ticket number",
            "travel_class": "Travel class (1 = first, 2 = second, 3 = third)",
            "fare": "Fare",
            "cabin": "Cabin number",
            "port_embarked": "Port of embarkation",
            "survived": "Whether the passenger survived the accident (0 = no, 1 = yes)",
        },
    )



import gzip
import os
import struct
import sys
import urllib.request
from array import array
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.error import HTTPError

import torch
from safeds._config import _init_default_device
from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
from safeds.data.labeled.containers import ImageDataset
from safeds.data.tabular.containers import Column

if TYPE_CHECKING:
    from safeds.data.image.containers import ImageList

_mnist_links: list[str] = ["http://yann.lecun.com/exdb/mnist/", "https://ossci-datasets.s3.amazonaws.com/mnist/"]
_mnist_files: dict[str, str] = {
    "train-images-idx3": "train-images-idx3-ubyte.gz",
    "train-labels-idx1": "train-labels-idx1-ubyte.gz",
    "test-images-idx3": "t10k-images-idx3-ubyte.gz",
    "test-labels-idx1": "t10k-labels-idx1-ubyte.gz",
}
_mnist_labels: dict[int, str] = {0: "0", 1: "1", 2: "2", 3: "3", 4: "4", 5: "5", 6: "6", 7: "7", 8: "8", 9: "9"}
_mnist_folder: str = "mnist"

_fashion_mnist_links: list[str] = ["http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/"]
_fashion_mnist_files: dict[str, str] = _mnist_files
_fashion_mnist_labels: dict[int, str] = {
    0: "T-shirt/top",
    1: "Trouser",
    2: "Pullover",
    3: "Dress",
    4: "Coat",
    5: "Sandal",
    6: "Shirt",
    7: "Sneaker",
    8: "Bag",
    9: "Ankle boot",
}
_fashion_mnist_folder: str = "fashion-mnist"

_kuzushiji_mnist_links: list[str] = ["http://codh.rois.ac.jp/kmnist/dataset/kmnist/"]
_kuzushiji_mnist_files: dict[str, str] = _mnist_files
_kuzushiji_mnist_labels: dict[int, str] = {
    0: "\u304a",
    1: "\u304d",
    2: "\u3059",
    3: "\u3064",
    4: "\u306a",
    5: "\u306f",
    6: "\u307e",
    7: "\u3084",
    8: "\u308c",
    9: "\u3092",
}
_kuzushiji_mnist_folder: str = "kmnist"


def load_mnist(path: str | Path, download: bool = True) -> tuple[ImageDataset[Column], ImageDataset[Column]]:
    """
    Load the `MNIST <http://yann.lecun.com/exdb/mnist/>`_ datasets.

    Parameters
    ----------
    path:
        the path were the files are stored or will be downloaded to
    download:
        whether the files should be downloaded to the given path

    Returns
    -------
    train_dataset, test_dataset:
        The train and test datasets.

    Raises
    ------
    FileNotFoundError
        if a file of the dataset cannot be found
    """
    path = Path(path) / _mnist_folder
    path.mkdir(parents=True, exist_ok=True)
    path_files = os.listdir(path)
    missing_files = []
    for file_path in _mnist_files.values():
        if file_path not in path_files:
            missing_files.append(file_path)
    if len(missing_files) > 0:
        if download:
            _download_mnist_like(
                path,
                {name: f_path for name, f_path in _mnist_files.items() if f_path in missing_files},
                _mnist_links,
            )
        else:
            raise FileNotFoundError(f"Could not find files {[str(path / file) for file in missing_files]}")
    return _load_mnist_like(path, _mnist_files, _mnist_labels)


def load_fashion_mnist(path: str | Path, download: bool = True) -> tuple[ImageDataset[Column], ImageDataset[Column]]:
    """
    Load the `Fashion-MNIST <https://github.com/zalandoresearch/fashion-mnist>`_ datasets.

    Parameters
    ----------
    path:
        the path were the files are stored or will be downloaded to
    download:
        whether the files should be downloaded to the given path

    Returns
    -------
    train_dataset, test_dataset:
        The train and test datasets.

    Raises
    ------
    FileNotFoundError
        if a file of the dataset cannot be found
    """
    path = Path(path) / _fashion_mnist_folder
    path.mkdir(parents=True, exist_ok=True)
    path_files = os.listdir(path)
    missing_files = []
    for file_path in _fashion_mnist_files.values():
        if file_path not in path_files:
            missing_files.append(file_path)
    if len(missing_files) > 0:
        if download:
            _download_mnist_like(
                path,
                {name: f_path for name, f_path in _fashion_mnist_files.items() if f_path in missing_files},
                _fashion_mnist_links,
            )
        else:
            raise FileNotFoundError(f"Could not find files {[str(path / file) for file in missing_files]}")
    return _load_mnist_like(path, _fashion_mnist_files, _fashion_mnist_labels)


def load_kmnist(path: str | Path, download: bool = True) -> tuple[ImageDataset[Column], ImageDataset[Column]]:
    """
    Load the `Kuzushiji-MNIST <https://github.com/rois-codh/kmnist>`_ datasets.

    Parameters
    ----------
    path:
        the path were the files are stored or will be downloaded to
    download:
        whether the files should be downloaded to the given path

    Returns
    -------
    train_dataset, test_dataset:
        The train and test datasets.

    Raises
    ------
    FileNotFoundError
        if a file of the dataset cannot be found
    """
    path = Path(path) / _kuzushiji_mnist_folder
    path.mkdir(parents=True, exist_ok=True)
    path_files = os.listdir(path)
    missing_files = []
    for file_path in _kuzushiji_mnist_files.values():
        if file_path not in path_files:
            missing_files.append(file_path)
    if len(missing_files) > 0:
        if download:
            _download_mnist_like(
                path,
                {name: f_path for name, f_path in _kuzushiji_mnist_files.items() if f_path in missing_files},
                _kuzushiji_mnist_links,
            )
        else:
            raise FileNotFoundError(f"Could not find files {[str(path / file) for file in missing_files]}")
    return _load_mnist_like(path, _kuzushiji_mnist_files, _kuzushiji_mnist_labels)


def _load_mnist_like(
    path: str | Path,
    files: dict[str, str],
    labels: dict[int, str],
) -> tuple[ImageDataset[Column], ImageDataset[Column]]:
    _init_default_device()

    path = Path(path)
    test_labels: Column | None = None
    train_labels: Column | None = None
    test_image_list: ImageList | None = None
    train_image_list: ImageList | None = None
    for file_name, file_path in files.items():
        if "idx1" in file_name:
            with gzip.open(path / file_path, mode="rb") as label_file:
                magic, size = struct.unpack(">II", label_file.read(8))
                if magic != 2049:
                    raise ValueError(f"Magic number mismatch. Actual {magic} != Expected 2049.")  # pragma: no cover
                if "train" in file_name:
                    train_labels = Column(
                        file_name,
                        [labels[label_index] for label_index in array("B", label_file.read())],
                    )
                else:
                    test_labels = Column(
                        file_name,
                        [labels[label_index] for label_index in array("B", label_file.read())],
                    )
        else:
            with gzip.open(path / file_path, mode="rb") as image_file:
                magic, size, rows, cols = struct.unpack(">IIII", image_file.read(16))
                if magic != 2051:
                    raise ValueError(f"Magic number mismatch. Actual {magic} != Expected 2051.")  # pragma: no cover
                image_data = array("B", image_file.read())
                image_tensor = torch.empty(size, 1, rows, cols, dtype=torch.uint8)
                for i in range(size):
                    image_tensor[i, 0] = torch.frombuffer(
                        image_data[i * rows * cols : (i + 1) * rows * cols],
                        dtype=torch.uint8,
                    ).reshape(rows, cols)
                image_list = _SingleSizeImageList()
                image_list._tensor = image_tensor
                image_list._tensor_positions_to_indices = list(range(size))
                image_list._indices_to_tensor_positions = image_list._calc_new_indices_to_tensor_positions()
                if "train" in file_name:
                    train_image_list = image_list
                else:
                    test_image_list = image_list
    if train_image_list is None or test_image_list is None or train_labels is None or test_labels is None:
        raise ValueError  # pragma: no cover
    return ImageDataset[Column](train_image_list, train_labels, 32, shuffle=True), ImageDataset[Column](
        test_image_list,
        test_labels,
        32,
    )


def _download_mnist_like(path: str | Path, files: dict[str, str], links: list[str]) -> None:
    path = Path(path)
    for file_name, file_path in files.items():
        for link in links:
            try:
                print(f"Trying to download file {file_name} via {link + file_path}")  # noqa: T201
                urllib.request.urlretrieve(link + file_path, path / file_path, reporthook=_report_download_progress)
                print()  # noqa: T201
                break
            except HTTPError as e:
                print(f"An error occurred while downloading: {e}")  # noqa: T201  # pragma: no cover


def _report_download_progress(current_packages: int, package_size: int, file_size: int) -> None:
    percentage = min(((current_packages * package_size) / file_size) * 100, 100)
    sys.stdout.write(f"\rDownloading... {percentage:.1f}%")
    sys.stdout.flush()
